import os
import ctypes
import numpy
from ctypes.util import find_library
from lib.guid import GUID
from enum import IntEnum

# DLL/SO/DYLIB library name; this will get a platform-dependent decoration ('lib' prefix on *nixes and appropriate platform-dependent extension)
__library_name__ = 'archipelago'

# enumerator of supported diagnostic levels (parameter to init_library call)
class Diagnostic_Level(IntEnum):
    No_Output = 0
    Errors = 1
    Warnings = 2
    Info = 3,
    Debug = 4
    Verbose = 5

# internal enumerator of library functions return codes
class Lib_Return_Code(IntEnum):
    OK = 0
    Unspecified_Error = 1
    Already_Initialized = 2
    Not_Initialized = 3
    Communicator_Error = 4
    Invalid_Argument = 5
    Not_Available = 6
    Invalid_State = 7

def get_lib_path(libname):
    if os.name == 'nt':
        return "%s\\%s" % (os.getcwd(), __library_name__)
    else:
        return ctypes.util.find_library(__library_name__)

# Archipelago library bridge
class Archipelago:
    def __init__(self):
        self.lib = None

    def __describe_error__(self, error_code):
        if error_code != Lib_Return_Code.OK:
            print("An error has occurred during archipelago execution:")

        if error_code == Lib_Return_Code.OK:
            pass
        elif error_code == Lib_Return_Code.Unspecified_Error:
            print("Unspecified error")
        elif error_code == Lib_Return_Code.Already_Initialized:
            print("The library or its component is already initialized")
        elif error_code == Lib_Return_Code.Not_Initialized:
            print("The library or its component is not initialized, or was deinitializer prior this call")
        elif error_code == Lib_Return_Code.Communicator_Error:
            print("The communicator reported an error")
        elif error_code == Lib_Return_Code.Invalid_Argument:
            print("Invalid argument given to function call")
        elif error_code == Lib_Return_Code.Not_Available:
            print("The requested resource is not available at the moment")
        elif error_code == Lib_Return_Code.Invalid_State:
            print("The call you requested cannot be performed in current state")

    # initializes archipelago library
    # the library needs to be present in the current working directory; it saves the current 
    # THREAD UNSAFE
    def init_library(self, diagnostics_level = Diagnostic_Level.No_Output):
        if self.lib == None:
            try:
                if 'LD_LIBRARY_PATH' in os.environ:
                    os.environ['LD_LIBRARY_PATH'] = os.getcwd() + ':' + os.environ['LD_LIBRARY_PATH']
                self.lib = ctypes.cdll.LoadLibrary(get_lib_path(__library_name__))
            except Exception as ex:
                print('Unable to load archipelago library (%s.dll, lib%s.so, lib%s.dylib). Please, make sure that the library is present in the current working directory and is a valid shared object for your OS type!' % (__library_name__, __library_name__, __library_name__))
                print(ex)
                return False

        if self.lib == None:
            return False

        # all the functions return c_int, which is an underlying type for NLib_Return_Code in native code

        # archipelago_init takes diagnostics level as an argument
        self.lib.archipelago_init.restype = ctypes.c_int
        self.lib.archipelago_start.argtypes = [ctypes.c_int]

        # archipelago_deinit has no arguments
        self.lib.archipelago_deinit.restype = ctypes.c_int

        # archipelago_start takes override host and port as its arguments
        self.lib.archipelago_start.restype = ctypes.c_int
        self.lib.archipelago_start.argtypes = [ctypes.c_char_p, ctypes.c_int]

        # archipelago_broadcast_candidate takes candidate-related fields as arguments (candidate array, its length, generation, source GUID)
        self.lib.archipelago_broadcast_candidate.restype = ctypes.c_int
        self.lib.archipelago_broadcast_candidate.argtypes = [ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64, ctypes.POINTER(GUID)]

        # archipelago_has_foreign_candidate takes bool pointer as its argument (flag indicating candidate presence)
        self.lib.archipelago_has_foreign_candidate.restype = ctypes.c_int
        self.lib.archipelago_has_foreign_candidate.argtypes = [ctypes.POINTER(ctypes.c_bool)]

        # archipelago_pop_foreign_candidate takes candidate-related field pointers as arguments (candidate array, its length, generation, source GUID)
        self.lib.archipelago_pop_foreign_candidate.restype = ctypes.c_int
        self.lib.archipelago_pop_foreign_candidate.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(GUID)]

        # archipelago_broadcast_best takes best solution-related fields as arguments (candidate array, its length, generation, source GUID)
        self.lib.archipelago_broadcast_best.restype = ctypes.c_int
        self.lib.archipelago_broadcast_best.argtypes = [ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64, ctypes.POINTER(GUID)]

        # archipelago_finalize has no arguments
        self.lib.archipelago_finalize.restype = ctypes.c_int

        # archipelago_has_foreign_best takes bool pointer as its argument (flag indicating best foreign solution presence)
        self.lib.archipelago_has_foreign_best.restype = ctypes.c_int
        self.lib.archipelago_has_foreign_best.argtypes = [ctypes.POINTER(ctypes.c_bool)]

        # archipelago_pop_foreign_best takes best solution-related field pointers as arguments (best solution array, its length, generation, source GUID)
        self.lib.archipelago_pop_foreign_best.restype = ctypes.c_int
        self.lib.archipelago_pop_foreign_best.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(GUID)]

        # archipelago_get_local_guid takes a pointer to GUID structure as its argument
        self.lib.archipelago_get_local_guid.restype = ctypes.c_int
        self.lib.archipelago_get_local_guid.argtypes = [ctypes.POINTER(GUID)]

        # initialize archipelago library
        result = self.lib.archipelago_init(ctypes.c_int(diagnostics_level.value))
        self.__describe_error__(result)
        return (result == Lib_Return_Code.OK)

    # deinitializes archipelago library
    # this releases all resources allocated by the library and allows for safe library release
    # THREAD UNSAFE
    def deinit_library(self, release_library = True):
        result = self.lib.archipelago_deinit()

        if release_library == True:
            del self.lib
            self.lib = None

        self.__describe_error__(result)
        return (result == Lib_Return_Code.OK)

    # starts the archipelago communication
    # this connect to all hosts and established the archipelago
    # THREAD UNSAFE
    def start_comm(self, bind_addr_override = "", bind_port_override = 0):
        addr = ctypes.create_string_buffer(bind_addr_override.encode("UTF-8"))
        result = self.lib.archipelago_start(addr, ctypes.c_int(int(bind_port_override)))
        self.__describe_error__(result)
        return (result == Lib_Return_Code.OK)

    # broadcasts candidate solution from local island to all other isles
    # candidate is a numpy vector
    # generation is non-decreasing sequence; if the sending/receiving buffer is full, oldest generation member is overwritten by new candidate
    # (thread-safe)
    def broadcast_candidate(self, candidate, generation, source_guid):
        result = self.lib.archipelago_broadcast_candidate(ctypes.c_void_p(candidate.ctypes.data), candidate.size, generation, ctypes.byref(source_guid))
        self.__describe_error__(result)

    # is there any foreign candidate waiting in the queue?
    # (thread-safe)
    def has_foreign_candidate(self):
        cflag = ctypes.c_bool()
        result = self.lib.archipelago_has_foreign_candidate(ctypes.byref(cflag))
        self.__describe_error__(result)
        return cflag.value

    # pops foreign candidate solution, if there is any
    # returns next foreign candidate solution (as numpy vector) and source GUID, if present
    # (thread-safe); but still can fail (return failure code), if other thread managed to withdraw the candidate between has_foreign_candidate and pop_foreign_candidate call
    def pop_foreign_candidate(self):

        # at first, create space for candidate length, generation and source GUID
        candidate_len = ctypes.c_uint64()
        candidate_gen = ctypes.c_uint64()
        src_guid = GUID()

        # retrieve length of the candidate
        result = self.lib.archipelago_pop_foreign_candidate(ctypes.c_void_p(None), ctypes.byref(candidate_len), ctypes.byref(candidate_gen), ctypes.byref(src_guid))
        self.__describe_error__(result)
        if result != 0:
            return False, 0, 0, 0

        # initialize candidate ndarray and retrieve the candidate contents
        candidate = numpy.zeros(candidate_len.value)
        result = self.lib.archipelago_pop_foreign_candidate(ctypes.c_void_p(candidate.ctypes.data), ctypes.byref(candidate_len), ctypes.byref(candidate_gen), ctypes.byref(src_guid))
        self.__describe_error__(result)
        if result != 0:
            return False, 0, 0, 0

        src_guid.update_cache(self.lib)

        return True, candidate, candidate_gen.value, src_guid

    # broadcasts best solution of current population
    # this method should be called prior finalize() and no further steps can be called upon this call (e.g., broadcast_candidate, has_foreign_candidate and pop_foreign_candidate cannot be called after this call)
    def broadcast_best(self, candidate, generation, source_guid):
        result = self.lib.archipelago_broadcast_best(ctypes.c_void_p(candidate.ctypes.data), candidate.size, generation, ctypes.byref(source_guid))
        self.__describe_error__(result)

    # finalizes the communication and waits for other islands to finish calculations - this method blocks, until all best solutions are gathered from all foreign islands
    # this method blocks until all the nodes finishes calculation; then, outer code should request foreign best solutions using has_foreign_best and pop_foreign_best methods
    # after this, all connections are terminated, but library remains in memory
    # (thread-safe)
    def finalize(self):
        result = self.lib.archipelago_finalize()
        self.__describe_error__(result)
        return (result == Lib_Return_Code.OK)

    # is there any foreign best solution?
    # (thread-safe)
    def has_foreign_best(self):
        cflag = ctypes.c_bool()
        result = self.lib.archipelago_has_foreign_best(ctypes.byref(cflag))
        self.__describe_error__(result)
        return cflag.value

    # pops foreign best solution, if there is any
    # returns next foreign best solution (as numpy vector) and source GUID, if present
    # (thread-safe); but still can fail (return failure code), if other thread managed to withdraw the candidate between has_foreign_best and pop_foreign_best call
    def pop_foreign_best(self):

        # at first, create space for candidate length, generation and source GUID
        candidate_len = ctypes.c_uint64()
        candidate_gen = ctypes.c_uint64()
        src_guid = GUID()

        # retrieve length of the candidate
        result = self.lib.archipelago_pop_foreign_best(ctypes.c_void_p(None), ctypes.byref(candidate_len), ctypes.byref(candidate_gen), ctypes.byref(src_guid))
        self.__describe_error__(result)
        if result != Lib_Return_Code.OK:
            return False, 0, 0, 0

        # initialize candidate ndarray and retrieve the candidate contents
        candidate = numpy.zeros(candidate_len.value)
        result = self.lib.archipelago_pop_foreign_best(ctypes.c_void_p(candidate.ctypes.data), ctypes.byref(candidate_len), ctypes.byref(candidate_gen), ctypes.byref(src_guid))
        self.__describe_error__(result)
        if result != Lib_Return_Code.OK:
            return False, 0, 0, 0

        src_guid.update_cache(self.lib)

        return True, candidate, candidate_gen.value, src_guid

    # retrieves local island GUID generated for the current island by the archipelago library
    # NOTE: library must be initialized and communication started prior this method call
    # (thread-safe)
    def get_local_guid(self):
        localguid = GUID()
        result = self.lib.archipelago_get_local_guid(ctypes.byref(localguid))
        self.__describe_error__(result)
        localguid.update_cache(self.lib)
        return localguid
